@@ -14,10 +14,11 @@ use crate::{
     types::{
         self,
         oid::{MAX_OID_FIRST_OCTET, MAX_OID_SECOND_OCTET},
-        Tag,
+        Constraints, Enumerated, Tag,
     },
     Decode,
 };
+use chrono::{DateTime, FixedOffset, NaiveDateTime, TimeZone};
 
 pub use self::{config::DecoderOptions, error::Error};
 
@@ -110,6 +111,204 @@ impl<'input> Decoder<'input> {
 
         Ok(result)
     }
+    /// Decode an object identifier from a byte slice in BER format.
+    /// Function is public to be used by other codecs.
+    pub fn decode_object_identifier_from_bytes(
+        data: &[u8],
+    ) -> Result<crate::types::ObjectIdentifier, Error> {
+        use num_traits::ToPrimitive;
+        let (mut contents, root_octets) =
+            parser::parse_base128_number(data).map_err(error::map_nom_err)?;
+        let the_number = root_octets
+            .to_u32()
+            .context(error::IntegerOverflowSnafu { max_width: 32u32 })?;
+        let first: u32;
+        let second: u32;
+        const MAX_OID_THRESHOLD: u32 = MAX_OID_SECOND_OCTET + 1;
+        if the_number > MAX_OID_FIRST_OCTET * MAX_OID_THRESHOLD + MAX_OID_SECOND_OCTET {
+            first = MAX_OID_FIRST_OCTET;
+            second = the_number - MAX_OID_FIRST_OCTET * MAX_OID_THRESHOLD;
+        } else {
+            second = the_number % MAX_OID_THRESHOLD;
+            first = (the_number - second) / MAX_OID_THRESHOLD;
+        }
+        let mut buffer = alloc::vec![first, second];
+
+        while !contents.is_empty() {
+            let (c, number) = parser::parse_base128_number(contents).map_err(error::map_nom_err)?;
+            contents = c;
+            buffer.push(
+                number
+                    .to_u32()
+                    .context(error::IntegerOverflowSnafu { max_width: 32u32 })?,
+            );
+        }
+        crate::types::ObjectIdentifier::new(buffer).context(error::InvalidObjectIdentifierSnafu)
+    }
+    /// Parse any GeneralizedTime string, allowing for any from ASN.1 definition
+    /// TODO, move to type itself?
+    pub fn parse_any_generalized_time_string(
+        string: alloc::string::String,
+    ) -> Result<types::GeneralizedTime, Error> {
+        // Reference https://obj-sys.com/asn1tutorial/node14.html
+        // If data contains ., 3 decimal places of seconds are expected
+        // If data contains explict Z, result is UTC
+        // If data contains + or -, explicit timezone is given
+        // If neither Z nor + nor -, purely local time is implied
+        let len = string.len();
+        // Helper function to deal with fractions and without timezone
+        let parse_without_timezone = |string: &str| -> Result<NaiveDateTime, Error> {
+            // Handle both decimal cases (dot . and comma , )
+            let string: &str = &string.replace(",", ".");
+            if string.contains('.') {
+                // Use chrono to parse the string every time, since we don't the know the number of decimals places
+                NaiveDateTime::parse_from_str(string, "%Y%m%d%H%.f")
+                    .or_else(|_| NaiveDateTime::parse_from_str(string, "%Y%m%d%H%M%.f"))
+                    .or_else(|_| NaiveDateTime::parse_from_str(string, "%Y%m%d%H%M%S%.f"))
+                    .map_err(|_| Error::InvalidDate)
+            } else {
+                let fmt_string = match string.len() {
+                    8 => "%Y%m%d",
+                    10 => "%Y%m%d%H",
+                    12 => "%Y%m%d%H%M",
+                    14 => "%Y%m%d%H%M%S",
+                    _ => "",
+                };
+                match fmt_string.len() {
+                    l if l > 0 => NaiveDateTime::parse_from_str(string, fmt_string)
+                        .map_err(|_| Error::InvalidDate),
+                    _ => Err(Error::InvalidDate),
+                }
+            }
+        };
+        if string.ends_with('Z') {
+            let naive = parse_without_timezone(&string[..len - 1])?;
+            return Ok(naive.and_utc().into());
+        }
+        // Check for timezone offset
+        if len > 5
+            && string
+                .chars()
+                .nth(len - 5)
+                .map_or(false, |c| c == '+' || c == '-')
+        {
+            let naive = parse_without_timezone(&string[..len - 5])?;
+            let sign = match string.chars().nth(len - 5) {
+                Some('+') => 1,
+                Some('-') => -1,
+                _ => {
+                    return Err(Error::InvalidDate);
+                }
+            };
+            let offset_hours = string
+                .chars()
+                .skip(len - 4)
+                .take(2)
+                .collect::<alloc::string::String>()
+                .parse::<i32>()
+                .map_err(|_| Error::InvalidDate)?;
+            let offset_minutes = string
+                .chars()
+                .skip(len - 2)
+                .take(2)
+                .collect::<alloc::string::String>()
+                .parse::<i32>()
+                .map_err(|_| Error::InvalidDate)?;
+            if offset_hours > 23 || offset_minutes > 59 {
+                return Err(Error::InvalidDate);
+            }
+            let offset = FixedOffset::east_opt(sign * (offset_hours * 3600 + offset_minutes * 60))
+                .ok_or(Error::InvalidDate)?;
+            return TimeZone::from_local_datetime(&offset, &naive)
+                .single()
+                .ok_or(Error::InvalidDate);
+        }
+
+        // Parse without timezone details
+        let naive = parse_without_timezone(&string)?;
+        Ok(naive.and_utc().into())
+    }
+    /// Enforce CER/DER restrictions defined in Section 11.7, strictly raise error on non-compliant
+    pub fn parse_canonical_generalized_time_string(
+        string: alloc::string::String,
+    ) -> Result<types::GeneralizedTime, Error> {
+        let len = string.len();
+        // Helper function to deal with fractions of seconds and without timezone
+        let parse_without_timezone = |string: &str| -> core::result::Result<NaiveDateTime, Error> {
+            let len = string.len();
+            if string.contains('.') {
+                // https://github.com/chronotope/chrono/issues/238#issuecomment-378737786
+                NaiveDateTime::parse_from_str(string, "%Y%m%d%H%M%S%.f")
+                    .map_err(|_| Error::InvalidDate)
+            } else if len == 14 {
+                NaiveDateTime::parse_from_str(string, "%Y%m%d%H%M%S")
+                    .map_err(|_| Error::InvalidDate)
+            } else {
+                // CER/DER encoding rules don't allow for timezone offset +/
+                // Or missing seconds/minutes/hours
+                // Or comma , instead of dot .
+                // Or local time without timezone
+                Err(Error::InvalidDate)
+            }
+        };
+        if string.ends_with('Z') {
+            let naive = parse_without_timezone(&string[..len - 1])?;
+            Ok(naive.and_utc().into())
+        } else {
+            Err(Error::InvalidDate)
+        }
+    }
+    /// Parse any UTCTime string, can be any from ASN.1 definition
+    /// TODO, move to type itself?
+    pub fn parse_any_utc_time_string(
+        string: alloc::string::String,
+    ) -> Result<types::UtcTime, Error> {
+        // When compared to GeneralizedTime, UTC time has no fractions.
+        let len = string.len();
+        // Largest string, e.g. "820102070000-0500".len() == 17
+        if len > 17 {
+            return Err(Error::InvalidDate);
+        }
+        let format = if string.contains('Z') {
+            if len == 11 {
+                "%y%m%d%H%MZ"
+            } else {
+                "%y%m%d%H%M%SZ"
+            }
+        } else if len == 15 {
+            "%y%m%d%H%M%z"
+        } else {
+            "%y%m%d%H%M%S%z"
+        };
+        match len {
+            11 | 13 => {
+                let naive = NaiveDateTime::parse_from_str(&string, format)
+                    .map_err(|_| Error::InvalidDate)?;
+                Ok(naive.and_utc())
+            }
+            15 | 17 => Ok(DateTime::parse_from_str(&string, format)
+                .map_err(|_| Error::InvalidDate)?
+                .into()),
+            _ => Err(Error::InvalidDate),
+        }
+    }
+
+    /// Enforce CER/DER restrictions defined in Section 11.8, strictly raise error on non-compliant
+    pub fn parse_canonical_utc_time_string(
+        string: alloc::string::String,
+    ) -> Result<types::UtcTime, Error> {
+        let len = string.len();
+        if string.ends_with('Z') {
+            let naive = match len {
+                13 => NaiveDateTime::parse_from_str(&string, "%y%m%d%H%M%SZ")
+                    .map_err(|_| Error::InvalidDate)?,
+                _ => Err(Error::InvalidDate)?,
+            };
+            Ok(naive.and_utc())
+        } else {
+            Err(Error::InvalidDate)
+        }
+    }
 }
 
 impl<'input> crate::Decoder for Decoder<'input> {
@@ -148,17 +347,22 @@ impl<'input> crate::Decoder for Decoder<'input> {
         })
     }
 
-    fn decode_enumerated(&mut self, tag: Tag) -> Result<types::Integer> {
-        self.decode_integer(tag)
+    fn decode_enumerated<E: Enumerated>(&mut self, tag: Tag) -> Result<E> {
+        E::from_discriminant(
+            self.decode_integer(tag, <_>::default())?
+                .try_into()
+                .map_err(Error::custom)?,
+        )
+        .ok_or_else(|| Error::custom("no valid discriminant"))
     }
 
-    fn decode_integer(&mut self, tag: Tag) -> Result<types::Integer> {
+    fn decode_integer(&mut self, tag: Tag, _: Constraints) -> Result<types::Integer> {
         Ok(types::Integer::from_signed_bytes_be(
             self.parse_primitive_value(tag)?.1,
         ))
     }
 
-    fn decode_octet_string(&mut self, tag: Tag) -> Result<Vec<u8>> {
+    fn decode_octet_string(&mut self, tag: Tag, _: Constraints) -> Result<Vec<u8>> {
         let (identifier, contents) = self.parse_value(tag)?;
 
         if identifier.is_primitive() {
@@ -213,42 +417,14 @@ impl<'input> crate::Decoder for Decoder<'input> {
     }
 
     fn decode_object_identifier(&mut self, tag: Tag) -> Result<crate::types::ObjectIdentifier> {
-        use num_traits::ToPrimitive;
         let contents = self.parse_primitive_value(tag)?.1;
-        let (mut contents, root_octets) =
-            parser::parse_base128_number(contents).map_err(error::map_nom_err)?;
-        let the_number = root_octets
-            .to_u32()
-            .context(error::IntegerOverflowSnafu { max_width: 32u32 })?;
-        let first: u32;
-        let second: u32;
-        const MAX_OID_THRESHOLD: u32 = MAX_OID_SECOND_OCTET + 1;
-        if the_number > MAX_OID_FIRST_OCTET * MAX_OID_THRESHOLD + MAX_OID_SECOND_OCTET {
-            first = MAX_OID_FIRST_OCTET;
-            second = the_number - MAX_OID_FIRST_OCTET * MAX_OID_THRESHOLD;
-        } else {
-            second = the_number % MAX_OID_THRESHOLD;
-            first = (the_number - second) / MAX_OID_THRESHOLD;
-        }
-        let mut buffer = alloc::vec![first, second];
-
-        while !contents.is_empty() {
-            let (c, number) = parser::parse_base128_number(contents).map_err(error::map_nom_err)?;
-            contents = c;
-            buffer.push(
-                number
-                    .to_u32()
-                    .context(error::IntegerOverflowSnafu { max_width: 32u32 })?,
-            );
-        }
-
-        crate::types::ObjectIdentifier::new(buffer).context(error::InvalidObjectIdentifierSnafu)
+        Self::decode_object_identifier_from_bytes(contents)
     }
 
-    fn decode_bit_string(&mut self, tag: Tag) -> Result<types::BitString> {
+    fn decode_bit_string(&mut self, tag: Tag, _: Constraints) -> Result<types::BitString> {
         let (input, bs) =
             self::parser::parse_encoded_value(&self.config, self.input, tag, |input| {
-                let unused_bits = if let Some(bits) = input.get(0).copied() {
+                let unused_bits = if let Some(bits) = input.first().copied() {
                     bits
                 } else {
                     return Ok(types::BitString::new());
@@ -261,61 +437,117 @@ impl<'input> crate::Decoder for Decoder<'input> {
                         if let Some(last) = buffer.last_mut() {
                             *last &= !((1 << bits) - 1);
                         }
-                        if buffer.last().map_or(false, |i| *i == 0) {
-                            buffer.pop();
-                        }
 
-                        let string = types::BitString::from_vec(buffer);
+                        let mut string = types::BitString::from_vec(buffer);
+                        let bit_length = string
+                            .len()
+                            .checked_sub(bits as usize)
+                            .ok_or(Error::InvalidBitString { bits })?;
+                        string.truncate(bit_length);
 
-                        if string.not_any() {
-                            Ok(types::BitString::new())
-                        } else {
-                            Ok(string)
-                        }
+                        Ok(string)
                     }
                     _ => Err(Error::InvalidBitString { bits: unused_bits }),
                 }
             })?;
 
         self.input = input;
-        if let Some((i, _)) = bs
-            .as_raw_slice()
-            .iter()
-            .enumerate()
-            .rev()
-            .find(|(_, v)| **v != 0)
-        {
-            Ok(types::BitString::from_vec(bs.as_raw_slice()[..=i].to_vec()))
-        } else {
-            Ok(bs)
-        }
+        Ok(bs)
     }
 
-    fn decode_utf8_string(&mut self, tag: Tag) -> Result<types::Utf8String> {
-        let vec = self.decode_octet_string(tag)?;
+    fn decode_visible_string(
+        &mut self,
+        tag: Tag,
+        constraints: Constraints,
+    ) -> Result<types::VisibleString, Self::Error> {
+        types::VisibleString::try_from(self.decode_octet_string(tag, constraints)?)
+            .map_err(Error::custom)
+    }
+
+    fn decode_ia5_string(
+        &mut self,
+        tag: Tag,
+        constraints: Constraints,
+    ) -> Result<types::Ia5String> {
+        types::Ia5String::try_from(self.decode_octet_string(tag, constraints)?)
+            .map_err(Error::custom)
+    }
+
+    fn decode_printable_string(
+        &mut self,
+        tag: Tag,
+        constraints: Constraints,
+    ) -> Result<types::PrintableString> {
+        types::PrintableString::try_from(self.decode_octet_string(tag, constraints)?)
+            .map_err(Error::custom)
+    }
+
+    fn decode_numeric_string(
+        &mut self,
+        tag: Tag,
+        constraints: Constraints,
+    ) -> Result<types::NumericString> {
+        types::NumericString::try_from(self.decode_octet_string(tag, constraints)?)
+            .map_err(Error::custom)
+    }
+
+    fn decode_teletex_string(
+        &mut self,
+        tag: Tag,
+        constraints: Constraints,
+    ) -> Result<types::TeletexString> {
+        types::TeletexString::try_from(self.decode_octet_string(tag, constraints)?)
+            .map_err(Error::custom)
+    }
+
+    fn decode_bmp_string(&mut self, _: Tag, _constraints: Constraints) -> Result<types::BmpString> {
+        todo!()
+    }
+
+    fn decode_utf8_string(
+        &mut self,
+        tag: Tag,
+        constraints: Constraints,
+    ) -> Result<types::Utf8String> {
+        let vec = self.decode_octet_string(tag, constraints)?;
         types::Utf8String::from_utf8(vec)
             .ok()
             .context(error::InvalidUtf8Snafu)
     }
 
+    fn decode_general_string(
+        &mut self,
+        tag: Tag,
+        constraints: Constraints,
+    ) -> Result<types::GeneralString> {
+        <_>::try_from(self.decode_octet_string(tag, constraints)?).map_err(Error::custom)
+    }
+
     fn decode_generalized_time(&mut self, tag: Tag) -> Result<types::GeneralizedTime> {
-        let string = self.decode_utf8_string(tag)?;
-        chrono::NaiveDateTime::parse_from_str(&string, "%Y%m%d%H%M%SZ")
-            .ok()
-            .context(error::InvalidDateSnafu)
-            .map(|date| types::GeneralizedTime::from_utc(date, chrono::FixedOffset::east(0)))
+        let string = self.decode_utf8_string(tag, <_>::default())?;
+        if self.config.encoding_rules.is_ber() {
+            Self::parse_any_generalized_time_string(string)
+        } else {
+            Self::parse_canonical_generalized_time_string(string)
+        }
     }
 
     fn decode_utc_time(&mut self, tag: Tag) -> Result<types::UtcTime> {
-        let string = self.decode_utf8_string(tag)?;
-        chrono::NaiveDateTime::parse_from_str(&string, "%y%m%d%H%M%SZ")
-            .ok()
-            .context(error::InvalidDateSnafu)
-            .map(|date| types::UtcTime::from_utc(date, chrono::Utc))
+        // Reference https://obj-sys.com/asn1tutorial/node15.html
+        let string = self.decode_utf8_string(tag, <_>::default())?;
+        if self.config.encoding_rules.is_ber() {
+            Self::parse_any_utc_time_string(string)
+        } else {
+            Self::parse_canonical_utc_time_string(string)
+        }
     }
 
-    fn decode_sequence_of<D: Decode>(&mut self, tag: Tag) -> Result<Vec<D>, Self::Error> {
-        self.decode_sequence(tag, |decoder| {
+    fn decode_sequence_of<D: Decode>(
+        &mut self,
+        tag: Tag,
+        _: Constraints,
+    ) -> Result<Vec<D>, Self::Error> {
+        self.parse_constructed_contents(tag, true, |decoder| {
             let mut items = Vec::new();
 
             while let Ok(item) = D::decode(decoder) {
@@ -326,8 +558,12 @@ impl<'input> crate::Decoder for Decoder<'input> {
         })
     }
 
-    fn decode_set_of<D: Decode + Ord>(&mut self, tag: Tag) -> Result<types::SetOf<D>, Self::Error> {
-        self.decode_sequence(tag, |decoder| {
+    fn decode_set_of<D: Decode + Ord>(
+        &mut self,
+        tag: Tag,
+        _: Constraints,
+    ) -> Result<types::SetOf<D>, Self::Error> {
+        self.parse_constructed_contents(tag, true, |decoder| {
             let mut items = types::SetOf::new();
 
             while let Ok(item) = D::decode(decoder) {
@@ -350,22 +586,84 @@ impl<'input> crate::Decoder for Decoder<'input> {
         self.parse_constructed_contents(tag, false, D::decode)
     }
 
-    fn decode_set<FIELDS, SET, F>(&mut self, tag: Tag, decode_fn: F) -> Result<SET, Self::Error>
+    fn decode_set<FIELDS, SET, D, F>(
+        &mut self,
+        tag: Tag,
+        _decode_fn: D,
+        field_fn: F,
+    ) -> Result<SET, Self::Error>
     where
-        SET: Decode,
+        SET: Decode + crate::types::Constructed,
         FIELDS: Decode,
+        D: Fn(&mut Self, usize, Tag) -> Result<FIELDS, Self::Error>,
         F: FnOnce(Vec<FIELDS>) -> Result<SET, Self::Error>,
     {
-        self.decode_sequence(tag, |decoder| {
+        self.parse_constructed_contents(tag, true, |decoder| {
             let mut fields = Vec::new();
 
             while let Ok(value) = FIELDS::decode(decoder) {
                 fields.push(value);
             }
 
-            (decode_fn)(fields)
+            (field_fn)(fields)
         })
     }
+
+    fn decode_optional<D: Decode>(&mut self) -> Result<Option<D>, Self::Error> {
+        if D::TAG == Tag::EOC {
+            Ok(D::decode(self).ok())
+        } else {
+            self.decode_optional_with_tag(D::TAG)
+        }
+    }
+
+    /// Decode an the optional value in a `SEQUENCE` or `SET` with `tag`.
+    /// Passing the correct tag is required even when used with codecs where
+    /// the tag is not present.
+    fn decode_optional_with_tag<D: Decode>(&mut self, tag: Tag) -> Result<Option<D>, Self::Error> {
+        Ok(D::decode_with_tag(self, tag).ok())
+    }
+
+    fn decode_optional_with_constraints<D: Decode>(
+        &mut self,
+        constraints: Constraints,
+    ) -> Result<Option<D>, Self::Error> {
+        Ok(D::decode_with_constraints(self, constraints).ok())
+    }
+
+    fn decode_optional_with_tag_and_constraints<D: Decode>(
+        &mut self,
+        tag: Tag,
+        constraints: Constraints,
+    ) -> Result<Option<D>, Self::Error> {
+        Ok(D::decode_with_tag_and_constraints(self, tag, constraints).ok())
+    }
+
+    fn decode_choice<D>(&mut self, _: Constraints) -> Result<D, Self::Error>
+    where
+        D: crate::types::DecodeChoice,
+    {
+        let (_, identifier) =
+            parser::parse_identifier_octet(self.input).map_err(error::map_nom_err)?;
+        D::from_tag(self, identifier.tag)
+    }
+
+    fn decode_extension_addition_with_constraints<D>(
+        &mut self,
+        // Constraints are irrelevant using BER
+        _: Constraints,
+    ) -> core::result::Result<Option<D>, Self::Error>
+    where
+        D: Decode,
+    {
+        <Option<D>>::decode(self)
+    }
+
+    fn decode_extension_addition_group<D: Decode + crate::types::Constructed>(
+        &mut self,
+    ) -> Result<Option<D>, Self::Error> {
+        <Option<D>>::decode(self)
+    }
 }
 
 #[cfg(test)]
@@ -406,8 +704,8 @@ mod tests {
 
     #[test]
     fn boolean() {
-        assert_eq!(true, decode::<bool>(&[0x01, 0x01, 0xff]).unwrap());
-        assert_eq!(false, decode::<bool>(&[0x01, 0x01, 0x00]).unwrap());
+        assert!(decode::<bool>(&[0x01, 0x01, 0xff]).unwrap());
+        assert!(!decode::<bool>(&[0x01, 0x01, 0x00]).unwrap());
     }
 
     #[test]
@@ -447,9 +745,9 @@ mod tests {
         data[2] = 0x01;
         data[3] = 0x01;
         data[4] = 0x01;
-        let mut bigint = types::Integer::from(1);
+        let mut bigint = num_bigint::BigInt::from(1);
         bigint <<= 2048;
-        assert_eq!(bigint, decode(&data).unwrap());
+        assert_eq!(bigint, decode::<num_bigint::BigInt>(&data).unwrap());
     }
 
     #[test]
@@ -470,8 +768,9 @@ mod tests {
 
     #[test]
     fn bit_string() {
-        let bitstring =
+        let mut bitstring =
             types::BitString::from_vec([0x0A, 0x3B, 0x5F, 0x29, 0x1C, 0xD0][..].to_owned());
+        bitstring.truncate(bitstring.len() - 4);
 
         let primitive_encoded: types::BitString =
             decode(&[0x03, 0x07, 0x04, 0x0A, 0x3B, 0x5F, 0x29, 0x1C, 0xD0][..]).unwrap();
@@ -512,6 +811,55 @@ mod tests {
         assert_eq!(name, decode::<String>(indefinite_constructed).unwrap());
     }
 
+    #[test]
+    fn utc_time() {
+        let time =
+            crate::types::GeneralizedTime::parse_from_str("991231235959+0000", "%y%m%d%H%M%S%z")
+                .unwrap();
+        // 991231235959Z
+        let has_z = &[
+            0x17, 0x0D, 0x39, 0x39, 0x31, 0x32, 0x33, 0x31, 0x32, 0x33, 0x35, 0x39, 0x35, 0x39,
+            0x5A,
+        ];
+        // 991231235959+0000
+        let has_noz = &[
+            0x17, 0x11, 0x39, 0x39, 0x31, 0x32, 0x33, 0x31, 0x32, 0x33, 0x35, 0x39, 0x35, 0x39,
+            0x2B, 0x30, 0x30, 0x30, 0x30,
+        ];
+        assert_eq!(
+            time,
+            decode::<chrono::DateTime::<chrono::Utc>>(has_z).unwrap()
+        );
+
+        assert_eq!(
+            time,
+            crate::der::decode::<crate::types::UtcTime>(has_z).unwrap()
+        );
+
+        assert_eq!(
+            time,
+            decode::<chrono::DateTime::<chrono::Utc>>(has_noz).unwrap()
+        );
+        assert!(crate::der::decode::<crate::types::UtcTime>(has_noz).is_err());
+    }
+
+    #[test]
+    fn generalized_time() {
+        let time = crate::types::GeneralizedTime::parse_from_str(
+            "20001231205959.999+0000",
+            "%Y%m%d%H%M%S%.3f%z",
+        )
+        .unwrap();
+        let has_z = &[
+            0x18, 0x13, 0x32, 0x30, 0x30, 0x30, 0x31, 0x32, 0x33, 0x31, 0x32, 0x30, 0x35, 0x39,
+            0x35, 0x39, 0x2E, 0x39, 0x39, 0x39, 0x5A,
+        ];
+        assert_eq!(
+            time,
+            decode::<chrono::DateTime::<chrono::FixedOffset>>(has_z).unwrap()
+        );
+    }
+
     #[test]
     fn sequence_of() {
         let vec = alloc::vec!["Jon", "es"];
@@ -535,14 +883,22 @@ mod tests {
             ok: bool,
         }
 
+        impl types::Constructed for Foo {
+            const FIELDS: types::fields::Fields = types::fields::Fields::from_static(&[
+                types::fields::Field::new_required(Ia5String::TAG, Ia5String::TAG_TREE),
+                types::fields::Field::new_required(bool::TAG, bool::TAG_TREE),
+            ]);
+        }
+
         impl types::AsnType for Foo {
             const TAG: Tag = Tag::SEQUENCE;
         }
 
         impl Decode for Foo {
-            fn decode_with_tag<D: crate::Decoder>(
+            fn decode_with_tag_and_constraints<D: crate::Decoder>(
                 decoder: &mut D,
                 tag: Tag,
+                _: Constraints,
             ) -> Result<Self, D::Error> {
                 decoder.decode_sequence(tag, |sequence| {
                     let name: Ia5String = Ia5String::decode(sequence)?;
@@ -553,7 +909,7 @@ mod tests {
         }
 
         let foo = Foo {
-            name: String::from("Smith").into(),
+            name: String::from("Smith").try_into().unwrap(),
             ok: true,
         };
         let bytes = &[
@@ -574,7 +930,7 @@ mod tests {
         type Type5 = Implicit<C2, Type2>;
 
         let jones = String::from("Jones");
-        let jones1 = Type1::from(jones);
+        let jones1 = Type1::try_from(jones).unwrap();
         let jones2 = Type2::from(jones1.clone());
         let jones3 = Type3::from(jones2.clone());
         let jones4 = Type4::from(jones3.clone());
